@@ -32,6 +32,7 @@ import (
 	"log"
 	"net/http"
 	"os"
+	"path/filepath"
 	"regexp"
 	"strings"
 	"time"
@@ -41,6 +42,7 @@ import (
 	"github.com/go-openapi/runtime"
 	"github.com/google/go-containerregistry/pkg/name"
 	v1 "github.com/google/go-containerregistry/pkg/v1"
+	ggcrlayout "github.com/google/go-containerregistry/pkg/v1/layout"
 	"github.com/google/go-containerregistry/pkg/v1/remote/transport"
 	"github.com/nozzle/throttler"
 	ssldsse "github.com/secure-systems-lab/go-securesystemslib/dsse"
@@ -331,16 +333,18 @@ func verifyOCISignature(ctx context.Context, verifier signature.Verifier, sig pa
 // certificate chains up to a trusted root using intermediate certificate chain coming from CheckOpts.
 // Optionally verifies the subject and issuer of the certificate.
 func ValidateAndUnpackCert(cert *x509.Certificate, co *CheckOpts) (signature.Verifier, error) {
-	return ValidateAndUnpackCertWithIntermediates(cert, co, co.IntermediateCerts)
+	verifier, _, err := ValidateAndUnpackCertWithIntermediates(cert, co, co.IntermediateCerts)
+	return verifier, err
 }
 
 // ValidateAndUnpackCertWithIntermediates creates a Verifier from a certificate. Verifies that the
 // certificate chains up to a trusted root using intermediate cert passed as separate argument.
-// Optionally verifies the subject and issuer of the certificate.
-func ValidateAndUnpackCertWithIntermediates(cert *x509.Certificate, co *CheckOpts, intermediateCerts *x509.CertPool) (signature.Verifier, error) {
+// Optionally verifies the subject and issuer of the certificate. Returns the chain built from the
+// certificate pools. Clients must verify the validity of this chain against a provided timestamp.
+func ValidateAndUnpackCertWithIntermediates(cert *x509.Certificate, co *CheckOpts, intermediateCerts *x509.CertPool) (signature.Verifier, []*x509.Certificate, error) {
 	verifier, err := signature.LoadVerifier(cert.PublicKey, crypto.SHA256)
 	if err != nil {
-		return nil, fmt.Errorf("invalid certificate found on signature: %w", err)
+		return nil, nil, fmt.Errorf("invalid certificate found on signature: %w", err)
 	}
 
 	// Handle certificates where the Subject Alternative Name is not set to a supported
@@ -363,31 +367,37 @@ func ValidateAndUnpackCertWithIntermediates(cert *x509.Certificate, co *CheckOpt
 	var chains [][]*x509.Certificate
 	if co.TrustedMaterial != nil {
 		if chains, err = verify.VerifyLeafCertificate(cert.NotBefore, cert, co.TrustedMaterial); err != nil {
-			return nil, err
+			return nil, nil, err
 		}
 	} else {
 		// If the trusted root is not available, use the verifiers from cosign (legacy).
 		chains, err = TrustedCert(cert, co.RootCerts, intermediateCerts)
 		if err != nil {
-			return nil, err
+			return nil, nil, err
 		}
 	}
 
+	// handle if chains has more than one chain - grab first and print message
+	if len(chains) > 1 {
+		fmt.Fprintf(os.Stderr, "**Info** Multiple valid certificate chains found. Selecting the first for further verification.\n")
+	}
+	chain := chains[0]
+
 	err = CheckCertificatePolicy(cert, co)
 	if err != nil {
-		return nil, err
+		return nil, nil, err
 	}
 
 	// If IgnoreSCT is set, skip the SCT check
 	if co.IgnoreSCT {
-		return verifier, nil
+		return verifier, chains[0], nil
 	}
 	contains, err := ContainsSCT(cert.Raw)
 	if err != nil {
-		return nil, err
+		return nil, nil, err
 	}
 	if !contains && len(co.SCT) == 0 {
-		return nil, &VerificationFailure{
+		return nil, nil, &VerificationFailure{
 			fmt.Errorf("certificate does not include required embedded SCT and no detached SCT was set"),
 		}
 	}
@@ -395,38 +405,33 @@ func ValidateAndUnpackCertWithIntermediates(cert *x509.Certificate, co *CheckOpt
 	// If trusted root is available and the SCT is embedded, use the verifiers from sigstore-go (preferred).
 	if co.TrustedMaterial != nil && contains {
 		if err := verify.VerifySignedCertificateTimestamp(chains, 1, co.TrustedMaterial); err != nil {
-			return nil, err
+			return nil, nil, err
 		}
-		return verifier, nil
+		return verifier, chain, nil
 	}
 
-	// handle if chains has more than one chain - grab first and print message
-	if len(chains) > 1 {
-		fmt.Fprintf(os.Stderr, "**Info** Multiple valid certificate chains found. Selecting the first to verify the SCT.\n")
+	if len(chain) < 2 {
+		return nil, nil, errors.New("certificate chain must contain at least a certificate and its issuer")
 	}
 	if contains {
-		if err := VerifyEmbeddedSCT(context.Background(), chains[0], co.CTLogPubKeys); err != nil {
-			return nil, err
+		if err := VerifyEmbeddedSCT(context.Background(), chain, co.CTLogPubKeys); err != nil {
+			return nil, nil, err
 		}
-		return verifier, nil
-	}
-	chain := chains[0]
-	if len(chain) < 2 {
-		return nil, errors.New("certificate chain must contain at least a certificate and its issuer")
+		return verifier, chain, nil
 	}
 	certPEM, err := cryptoutils.MarshalCertificateToPEM(chain[0])
 	if err != nil {
-		return nil, err
+		return nil, nil, err
 	}
 	chainPEM, err := cryptoutils.MarshalCertificatesToPEM(chain[1:])
 	if err != nil {
-		return nil, err
+		return nil, nil, err
 	}
 	if err := VerifySCT(context.Background(), certPEM, chainPEM, co.SCT, co.CTLogPubKeys); err != nil {
-		return nil, err
+		return nil, nil, err
 	}
 
-	return verifier, nil
+	return verifier, chain, nil
 }
 
 // CheckCertificatePolicy checks that the certificate subject and issuer match
@@ -850,6 +855,7 @@ func verifyInternal(ctx context.Context, sig oci.Signature, h v1.Hash,
 	}
 
 	verifier := co.SigVerifier
+	var verifierChain []*x509.Certificate
 	if verifier == nil {
 		// If we don't have a public key to check against, we can try a root cert.
 		cert, err := sig.Cert()
@@ -881,10 +887,12 @@ func verifyInternal(ctx context.Context, sig oci.Signature, h v1.Hash,
 		if pool == nil {
 			pool = co.IntermediateCerts
 		}
-		verifier, err = ValidateAndUnpackCertWithIntermediates(cert, co, pool)
+		verifier, chain, err = ValidateAndUnpackCertWithIntermediates(cert, co, pool)
 		if err != nil {
 			return false, err
 		}
+		// Remove the end-entity certificate from the chain, as that will come from sig.Cert()
+		verifierChain = chain[1:]
 	}
 
 	// 1. Perform cryptographic verification of the signature using the certificate's public key.
@@ -910,14 +918,14 @@ func verifyInternal(ctx context.Context, sig oci.Signature, h v1.Hash,
 
 		if acceptableRFC3161Time != nil {
 			// Verify the cert against the timestamp time.
-			if err := CheckExpiry(cert, *acceptableRFC3161Time); err != nil {
+			if err := CheckExpiry(cert, verifierChain, *acceptableRFC3161Time); err != nil {
 				return false, fmt.Errorf("checking expiry on certificate with timestamp: %w", err)
 			}
 			expirationChecked = true
 		}
 
 		if acceptableRekorBundleTime != nil {
-			if err := CheckExpiry(cert, *acceptableRekorBundleTime); err != nil {
+			if err := CheckExpiry(cert, verifierChain, *acceptableRekorBundleTime); err != nil {
 				return false, fmt.Errorf("checking expiry on certificate with bundle: %w", err)
 			}
 			expirationChecked = true
@@ -925,7 +933,7 @@ func verifyInternal(ctx context.Context, sig oci.Signature, h v1.Hash,
 
 		// if no timestamp has been provided, use the current time
 		if !expirationChecked {
-			if err := CheckExpiry(cert, time.Now()); err != nil {
+			if err := CheckExpiry(cert, verifierChain, time.Now()); err != nil {
 				// If certificate is expired and not signed timestamp was provided then error the following message. Otherwise throw an expiration error.
 				if co.IgnoreTlog && acceptableRFC3161Time == nil {
 					return false, &VerificationFailure{
@@ -1062,6 +1070,11 @@ func VerifyLocalImageAttestations(ctx context.Context, path string, co *CheckOpt
 		return nil, false, errors.New("one of verifier, root certs, or trusted root is required")
 	}
 
+	// Check for v3 bundles first (if NewBundleFormat is enabled)
+	if co.NewBundleFormat {
+		return verifyLocalImageAttestationsSigstoreBundle(ctx, path, co)
+	}
+
 	se, err := layout.SignedImageIndex(path)
 	if err != nil {
 		return nil, false, err
@@ -1169,21 +1182,36 @@ func VerifyImageAttestation(ctx context.Context, atts oci.Signatures, h v1.Hash,
 	return checkedAttestations, bundleVerified, nil
 }
 
-// CheckExpiry confirms the time provided is within the valid period of the cert
-func CheckExpiry(cert *x509.Certificate, it time.Time) error {
+// CheckExpiry confirms the time provided is within the valid period of the certificate and optionally
+// all issuing CA certificates.
+func CheckExpiry(cert *x509.Certificate, issuingChain []*x509.Certificate, it time.Time) error {
 	ft := func(t time.Time) string {
 		return t.Format(time.RFC3339)
 	}
 	if cert.NotAfter.Before(it) {
 		return &VerificationFailure{
-			fmt.Errorf("certificate expired before signatures were entered in log: %s is before %s",
+			fmt.Errorf("certificate expired before observed time: %s is before %s",
 				ft(cert.NotAfter), ft(it)),
 		}
 	}
 	if cert.NotBefore.After(it) {
 		return &VerificationFailure{
-			fmt.Errorf("certificate was issued after signatures were entered in log: %s is after %s",
-				ft(cert.NotAfter), ft(it)),
+			fmt.Errorf("certificate was issued after observed time: %s is after %s",
+				ft(cert.NotBefore), ft(it)),
+		}
+	}
+	for _, c := range issuingChain {
+		if c.NotAfter.Before(it) {
+			return &VerificationFailure{
+				fmt.Errorf("issuing CA certificate expired before observed time: %s is before %s",
+					ft(c.NotAfter), ft(it)),
+			}
+		}
+		if c.NotBefore.After(it) {
+			return &VerificationFailure{
+				fmt.Errorf("issuing CA certificate was issued after observed time: %s is after %s",
+					ft(c.NotBefore), ft(it)),
+			}
 		}
 	}
 	return nil
@@ -1213,11 +1241,16 @@ func VerifyBundle(sig oci.Signature, co *CheckOpts) (bool, error) {
 		return false, errors.New("no trusted rekor public keys provided")
 	}
 
-	if err := compareSigs(bundle.Payload.Body.(string), sig); err != nil {
+	bundleBody, ok := bundle.Payload.Body.(string)
+	if !ok {
+		return false, errors.New("bundle payload body is not a string")
+	}
+
+	if err := compareSigs(bundleBody, sig); err != nil {
 		return false, err
 	}
 
-	if err := comparePublicKey(bundle.Payload.Body.(string), sig, co); err != nil {
+	if err := comparePublicKey(bundleBody, sig, co); err != nil {
 		return false, err
 	}
 
@@ -1230,7 +1263,7 @@ func VerifyBundle(sig oci.Signature, co *CheckOpts) (bool, error) {
 		return false, fmt.Errorf("reading base64signature: %w", err)
 	}
 
-	alg, bundlehash, err := bundleHash(bundle.Payload.Body.(string), signature)
+	alg, bundlehash, err := bundleHash(bundleBody, signature)
 	if err != nil {
 		return false, fmt.Errorf("computing bundle hash: %w", err)
 	}
@@ -1249,7 +1282,7 @@ func VerifyBundle(sig oci.Signature, co *CheckOpts) (bool, error) {
 		if err != nil {
 			return false, fmt.Errorf("decoding log ID: %w", err)
 		}
-		body, _ := base64.StdEncoding.DecodeString(payload.Body.(string))
+		body, _ := base64.StdEncoding.DecodeString(bundleBody)
 		entry, err := tlog.NewEntry(body, payload.IntegratedTime, payload.LogIndex, logID, bundle.SignedEntryTimestamp, nil)
 		if err != nil {
 			return false, fmt.Errorf("converting tlog entry: %w", err)
@@ -1322,6 +1355,9 @@ func (s *sigContent) MessageSignatureContent() verify.MessageSignatureContent {
 // returns the timestamp value.
 // It returns (nil, nil) if there is no timestamp, or (nil, err) if there is an invalid timestamp or if
 // no root is provided with a timestamp.
+//
+// Note: This function does not perform CRL/OCSP certificate revocation checks.
+// Callers are responsible for validating the TSA certificate and trusted material provided via CheckOpts according to their policy.
 func VerifyRFC3161Timestamp(sig oci.Signature, co *CheckOpts) (*timestamp.Timestamp, error) {
 	ts, err := sig.RFC3161Timestamp()
 	switch {
@@ -1676,6 +1712,146 @@ func GetBundles(_ context.Context, signedImgRef name.Reference, registryClientOp
 	return bundles, &h, nil
 }
 
+// bundleDescriptor holds the digest and path to a bundle blob in a local OCI layout
+type bundleDescriptor struct {
+	digest   v1.Hash
+	blobPath string
+}
+
+// HasLocalBundles checks if a local OCI layout has v3 sigstore bundles.
+// V3 bundles are stored as separate images with layers having
+// media type "application/vnd.dev.sigstore.bundle".
+func HasLocalBundles(path string) (bool, error) {
+	return hasLocalSigstoreBundles(path)
+}
+
+// HasLocalAttestationBundles checks if a local OCI layout has v3 sigstore bundles for attestations.
+// For v3, both signatures and attestations use the same bundle format.
+func HasLocalAttestationBundles(path string) (bool, error) {
+	return hasLocalSigstoreBundles(path)
+}
+
+// GetLocalBundles retrieves v3 sigstore bundles from a local OCI layout.
+// Returns bundles, target image hash, and error. Invalid bundles are logged and skipped.
+func GetLocalBundles(path string) ([]*sgbundle.Bundle, *v1.Hash, error) {
+	descriptors, hash, err := getLocalBundleDescriptors(path)
+	if err != nil {
+		return nil, nil, err
+	}
+
+	bundles := make([]*sgbundle.Bundle, 0, len(descriptors))
+	for _, descriptor := range descriptors {
+		bundleBytes, err := os.ReadFile(filepath.Join(path, descriptor.blobPath))
+		if err != nil {
+			ui.Warnf(context.Background(), "Failed to read bundle blob %s: %v", descriptor.digest.Hex, err)
+			continue
+		}
+
+		bundle := &sgbundle.Bundle{}
+		if err := bundle.UnmarshalJSON(bundleBytes); err != nil {
+			ui.Warnf(context.Background(), "Failed to unmarshal bundle %s: %v", descriptor.digest.Hex, err)
+			continue
+		}
+
+		if !bundle.MinVersion("v0.3") {
+			ui.Warnf(context.Background(), "Bundle %s version too old (requires v0.3+)", descriptor.digest.Hex)
+			continue
+		}
+
+		bundles = append(bundles, bundle)
+	}
+
+	if len(bundles) == 0 {
+		return nil, nil, &ErrNoMatchingAttestations{
+			fmt.Errorf("no valid bundles found in local layout"),
+		}
+	}
+
+	return bundles, hash, nil
+}
+
+func hasLocalSigstoreBundles(path string) (bool, error) {
+	descriptors, _, err := getLocalBundleDescriptors(path)
+	if err != nil {
+		return false, err
+	}
+	return len(descriptors) > 0, nil
+}
+
+func getLocalBundleDescriptors(path string) ([]bundleDescriptor, *v1.Hash, error) {
+	p, err := ggcrlayout.FromPath(path)
+	if err != nil {
+		return nil, nil, fmt.Errorf("loading OCI layout from %s: %w", path, err)
+	}
+
+	ii, err := p.ImageIndex()
+	if err != nil {
+		return nil, nil, fmt.Errorf("getting image index: %w", err)
+	}
+
+	manifest, err := ii.IndexManifest()
+	if err != nil {
+		return nil, nil, fmt.Errorf("getting index manifest: %w", err)
+	}
+
+	// Find the target image digest from the index manifest
+	var targetDigest v1.Hash
+	for _, m := range manifest.Manifests {
+		if val, ok := m.Annotations["kind"]; ok && val == "dev.cosignproject.cosign/image" {
+			targetDigest = m.Digest
+			break
+		}
+	}
+	if targetDigest.String() == "" {
+		return nil, nil, nil
+	}
+
+	// Scan blobs/sha256 directory for referrer manifests
+	blobsDir := filepath.Join(path, "blobs", "sha256")
+	entries, err := os.ReadDir(blobsDir)
+	if err != nil {
+		if errors.Is(err, os.ErrNotExist) {
+			return nil, nil, nil
+		}
+		return nil, nil, fmt.Errorf("reading blobs directory: %w", err)
+	}
+
+	var descriptors []bundleDescriptor
+	for _, entry := range entries {
+		if entry.IsDir() {
+			continue
+		}
+		blobPath := filepath.Join(blobsDir, entry.Name())
+		data, err := os.ReadFile(blobPath)
+		if err != nil {
+			continue
+		}
+
+		// Try to parse as manifest (skip blobs that aren't manifests)
+		blobManifest, err := v1.ParseManifest(bytes.NewReader(data))
+		if err != nil {
+			continue
+		}
+
+		// Check if this is a referrer manifest pointing to our target
+		if blobManifest.Subject != nil && blobManifest.Subject.Digest == targetDigest {
+			// Collect bundle layer descriptors from this referrer manifest
+			for _, layer := range blobManifest.Layers {
+				if strings.HasPrefix(string(layer.MediaType), "application/vnd.dev.sigstore.bundle") {
+					// layer.Digest.Hex is validated by go-containerregistry to be a valid hex hash,
+					// but use filepath.Clean for defense-in-depth against path traversal
+					descriptors = append(descriptors, bundleDescriptor{
+						digest:   layer.Digest,
+						blobPath: filepath.Clean(filepath.Join("blobs", "sha256", layer.Digest.Hex)),
+					})
+				}
+			}
+		}
+	}
+
+	return descriptors, &targetDigest, nil
+}
+
 // verifyImageAttestationsSigstoreBundle verifies attestations from attached sigstore bundles
 func verifyImageAttestationsSigstoreBundle(ctx context.Context, signedImgRef name.Reference, co *CheckOpts, nameOpts ...name.Option) (checkedAttestations []oci.Signature, atLeastOneBundleVerified bool, err error) {
 	bundles, hash, err := GetBundles(ctx, signedImgRef, co.RegistryClientOpts, nameOpts...)
@@ -1765,3 +1941,72 @@ func verifyImageAttestationsSigstoreBundle(ctx context.Context, signedImgRef nam
 
 	return checkedAttestations, atLeastOneBundleVerified, nil
 }
+
+// verifyLocalImageAttestationsSigstoreBundle verifies attestations from local sigstore bundles
+func verifyLocalImageAttestationsSigstoreBundle(ctx context.Context, path string, co *CheckOpts) (checkedAttestations []oci.Signature, bundleVerified bool, err error) {
+	bundles, hash, err := GetLocalBundles(path)
+	if err != nil {
+		return nil, false, err
+	}
+
+	digestBytes, err := hex.DecodeString(hash.Hex)
+	if err != nil {
+		return nil, false, err
+	}
+
+	artifactPolicyOption := verify.WithArtifactDigest(hash.Algorithm, digestBytes)
+
+	// For local bundles, we verify sequentially (local I/O is fast, no need for parallel throttler)
+	var atLeastOneBundleVerified bool
+	var errs []error
+	for _, bundle := range bundles {
+		_, err := VerifyNewBundle(ctx, co, artifactPolicyOption, bundle)
+		if err != nil {
+			// Log error and accumulate for final error message
+			errs = append(errs, err)
+			ui.Warnf(ctx, "Failed to verify bundle: %v", err)
+			continue
+		}
+
+		dsse, ok := bundle.Content.(*protobundle.Bundle_DsseEnvelope)
+		if !ok {
+			err := fmt.Errorf("bundle does not contain a DSSE envelope")
+			errs = append(errs, err)
+			ui.Warnf(ctx, "%v", err)
+			continue
+		}
+
+		payload, err := json.Marshal(dsse.DsseEnvelope)
+		if err != nil {
+			errs = append(errs, fmt.Errorf("marshaling DSSE envelope: %w", err))
+			ui.Warnf(ctx, "Failed to marshal DSSE envelope: %v", err)
+			continue
+		}
+
+		att, err := static.NewAttestation(payload)
+		if err != nil {
+			errs = append(errs, fmt.Errorf("creating attestation: %w", err))
+			ui.Warnf(ctx, "Failed to create attestation: %v", err)
+			continue
+		}
+
+		if co.ClaimVerifier != nil {
+			if err := co.ClaimVerifier(att, *hash, co.Annotations); err != nil {
+				errs = append(errs, fmt.Errorf("claim verification: %w", err))
+				ui.Warnf(ctx, "Claim verification failed: %v", err)
+				continue
+			}
+		}
+
+		checkedAttestations = append(checkedAttestations, att)
+		atLeastOneBundleVerified = true
+	}
+
+	if len(checkedAttestations) == 0 {
+		return nil, false, &ErrNoMatchingAttestations{
+			fmt.Errorf("no matching attestations: %w", errors.Join(errs...)),
+		}
+	}
+
+	return checkedAttestations, atLeastOneBundleVerified, nil
+}
